package PureMessage::Sieve;

use strict;

our $VERSION = '4.502';

use base 'Exporter';
our @EXPORT_OK = qw(sieve_parse_file sieve_print sieve_text sieve_node);

sub sieve_parse_file {
    require PureMessage::Sieve::Parser;
    my $p = PureMessage::Sieve::Parser->new;
    return $p->parse_file(@_);
}

sub sieve_print {
    require PureMessage::Sieve::Print;
    PureMessage::Sieve::Print::print(@_);
}

sub sieve_text {
    require PureMessage::Sieve::Print;
    PureMessage::Sieve::Print::text(@_);
}

sub sieve_node {
    require PureMessage::Sieve::Nodes;
    PureMessage::Sieve::Nodes::make(@_);
}

1;

__END__

=head1 NAME

PureMessage::Sieve - Sieve parsing and printing

=head1 SYNOPSIS

 use PureMessage::Sieve qw(sieve_parse_file sieve_print sieve_node);

 my $script = sieve_parse_file("policy.siv");
 print_sieve($script, *STDOUT);

=head1 DESCRIPTION

Sieve is a Mail Filtering Language.  Its specification is found in RFC
3028.

This module provide functions to convert textual Sieve into a syntax
tree and syntax trees back into textual Sieve.  The syntax tree is
built of nodes of different types.  The nodes provide methods to
inspect and modify the state of the nodes and the shape of the tree.

The following functions are provided.  None of them are exported by
default:

=over

=item $script = sieve_parse_file( $filename )

This will parse the Sieve source from the given file and return the
root of a syntax tree representing the script.  The function will
croak if the file can't be opened or if there is syntax errors in the
script.

=item sieve_print( $node )

=item sieve_print( $node, $filehandle )

=item sieve_print( $node, %opts )

This will format and print the Sieve syntax tree node given by $node
on the given filehandle.  The filehandle defaults to C<STDOUT> if not
specified.

Key/value style options can be provided to specify how the source
should be formatted.  The following options are recognized:

=over

=item line_width

How long are the lines allowed to be.  The default is 70.  Note that
long strings are not broken up so some lines might actually end up
longer.

=item indent_offset

How much indentation to use.  The default is 4.

=item comment

Provide a text string that is formatted as a comment in front of
the printed text.

=item file

Can be used to specify the file handle to print to when providing
options.

=back

No return value is provided.  Successful printing can be checked by
closing the handle and checking the return value from close().

=item $str = sieve_text( $node, %opts )

This will format the $node as textual Sieve and return it.  The
options are the same as for sieve_print().

=item $node = sieve_node( $type, @args )

Node constructor function. See the description of nodes below for how
to use if for different types.

=back

=head1 SYNTAX TREE NODES

The Sieve syntax tree is built up of nodes.  This section describe
relationship between nodes and the methods available for nodes of
various types.

Consider the following program:

   if size :over 1M {
      reject "Too big";
      stop;
   }
   elsif anyof(true, false) {
       keep;
   }
   else {
      discard;
   }

As a syntax tree it looks like this:

   script
    - commands
        \_ if
            - args
               \_ cond
               |   - test "size"
               |      - args
               |         \_ tag ":over"
               |         \_ number "1000000"
               |   - commands
               |      \_ action "reject"
               |      |   - args
               |      |      \_ string "Too big"
               |      \_ action "stop"
               \_ cond
               |   - test "anyof"
               |      - args
               |         \_ testlist
               |             - tests
               |                \_ test "true"
               |                \_ test "false"
               |   - commands
               |      \_ action "keep"
 	       \ cond
                   - test undef
		   - commands
                      \_ action "discard"

The root of a Sieve syntax tree is always a C<script> node.  The
C<script> node has the attribute C<commands> where all the commands of
the script hangs off.  In this case there is only a single command
node; C<if>.  An C<if> node has an attribute called C<args> which
contains C<cond> nodes.  In this case there is 3 C<cond> nodes that
each have the attribute C<test> and C<commands>.

The node types are themselves are organized in an hierachy.  This
diagram shows the node type hierarchy:

   node
    \_ script
    \_ command
    |   \_ test
    |   \_ action
    |   \_ if
    \_ cond
    \_ testlist
    \_ stringlist
    \_ string
    \_ tag
    \_ number

The following method is provided by all nodes:

=over

=item $node->type

This returns a string denoting the type of this node.  The type is a
short all lower case string like C<script>, C<if>, C<string>.  It is not
possible to modify the type of a node.

=back

=head2 script

This kind of node is used as the root of a complete Sieve syntax tree.
The following methods are provided for script nodes:

=over

=item @req = $script->requires

=item $script->requires( @new_req )

This get/set the list of strings that denote the required extensions
to execute this script.  Note that the elements of the requires list
is plain perl strings, not Sieve string nodes.

=item @cmd = $script->commands

=item $script->commands( @new_cmd )

This get/set the list of commands that make up the script.  The
elements will be either C<action> or C<if> nodes.

=back

A script node is constructed with:

=over

=item $node = sieve_node( "script" )

=item $node = sieve_node( "script", \@req, \@cmd )

This constructs a new script node containing the given requires and
commands.

=back

=head2 command

These kind of nodes represent the commands of a Sieve script.  The
following methods are provided:

=over

=item $name = $cmd->name

=item $cmd->name( $new_name )

This get/set the name of the command.  The name is a plain string like
"if", "true", "fileinto" or "reject".

=item @arg = $cmd->args

=item $cmd->args( @new_arg )

This get/set the arguments of the command.  Arguments comes in many
different types.  Use the $arg->type method to determine which one.

=item $com = $cmd->comment

=item $cmd->comment( $new_com )

This get/set the comment assosiated with the command.  The comment is
a plain, and possibly multilined, perl string.

=item $line = $cmd->line

=item $cmd->line( $new_line )

This get/set the line number where the command occurs in the script
file.  Line numbers start at 1.

=back

Command nodes are not constructed directly.  Instead you should
construct one of the subtypes; C<test>, C<action> or C<if>.

=head2 test and action

Tests and action nodes are specializations of the command type.  They
provide no new methods.

Tests are different in that their argument can be other C<test>s or
C<testlist>s.

These nodes are constructed with:

=over

=item $test = sieve_node( "test", $name, @arg )

This construct a new test node.

=item $action = sieve_node( "action", $name, @arg )

This construct a new action node.

=back


=head2 if

A complete if/elsif/else block is represented by an C<if> node.

An C<if> node is a specialization of the command type.  It provides no
new methods, but the args of an C<if> node are all C<cond> nodes.
Each cond node represent a single C<if>, C<elsif> or C<else>.

An C<if> node is constructed with:

=over

=item $cond = sieve_node( "if", "if", @cond )

This construct a new if node.

=back

=head2 cond

A single conditional test within an if/elsif/else block is represented
by a C<cond> node.  The following methods are provided:

=over

=item $test = $cond->test

=item $cond->test( $new_test )

This get/set the test assosicated with this block.  The last cond
child of an C<if> node might have no test and will then represent an
C<else> block.

=item $com = $cond->comment

=item $cond->comment( $new_com )

This get/set the assosiated comment.  Same behaviour as the comment
method of C<command> nodes.

=item @cmd = $cond->commands

=item $cond->commands( @new_cmd )

This get/set the assosicated commands.  Same behaviour as the commands
method of C<script> nodes.

=back

A C<cond> node is constructed with:

=over

=item $cond = sieve_node( "cond", $test, $comment, @cmd )

This construct a new cond node.

=back

=head2 testlist

This kind of node represent a list of tests.  It can only occur as an
argument of a C<test> node and will then always be the only argument.

The following additonal method is provided:

=over

=item @test = $list->tests

=item $list->tests( @new_test )

This get/set the tests contained in the list.

=back

A testlist node is constructed with:

=over

=item $node = sieve_node( "testlist", @test )

This constructs a new testlist node containing the given tests.

=back

=head2 stringlist

This kind of node represent a list of strings.  The following
additonal method is provided:

=over

=item @str = $list->strings

=item $list->strings( @new_str )

This get/set the strings contained in the list.  When setting plain
perl strings might be provided and they are automatically converted to
string nodes.

=back

A stringlist node is constructed with:

=over

=item $node = sieve_node( "stringlist", @str )

This constructs a new stringlist node containing the given strings.

=back

=head2 string, tag and number

These nodes represent the leaves of the syntax tree.  These nodes
provide no additional methods, but use overloading so that if they can
be used as plain perl scalars.  Note that C<number> nodes might
provide a different value in string and numeric context.

These nodes are constructed with:

=over

=item $node = sieve_node( $type, $value )

This constructs a new leaf node of the given type; which should be one
of C<string>, C<tag> or C<number>.  The $value for C<string> can be
anything.  The $value for C<tag> must start with ":" and then contain
one or more identifier letters.  The $value for C<number> must be a
decimal number (note that it cannot be a string like "1K" even if this
is legal Sieve source representation of a number).

Examples:

  $string = sieve_node("string", "Mary had a little lamb.");
  $tag = sieve_node("tag", ":contains");
  $number = sieve_node("number", 1024);

=back

=head1 SEE ALSO

RFC 3028
