package PureMessage::Sieve::Nodes;

use strict;

use base 'Exporter';

BEGIN {
    our @EXPORT = ();
    my @code;
    for (qw(Script Test Action If Cond Testlist Stringlist String Tag Number)) {
	my $mk = "new_\L$_";
	push(@code, "sub $mk { 'PureMessage::Sieve::$_'->new(\@_) }\n");
	push(@EXPORT, $mk);
    }
    #print join("", @code), "\n";
    eval join("", @code);
    die $@ if $@;
}

sub make {
    my $type = shift;
    my $class = "PureMessage::Sieve::" . ucfirst($type);
    $class->new(@_);
}

package PureMessage::Sieve::Node;

sub type {
    my $self = shift;
    my $type = ref($self);
    $type =~ s/^PureMessage::Sieve:://;
    lc($type);
}

sub _flatten {
    my $self = shift;
    map { ref($_) eq 'ARRAY' ? @$_ : $_ } @_;
}

package PureMessage::Sieve::Script;

our @ISA = qw(PureMessage::Sieve::Node);

sub new {
    my($class, $requires, $commands) = @_;
    $requires ||= [];
    $commands ||= [];
    bless [$requires, $commands], $class;
}

sub requires {
    my $self = shift;
    return @{$self->[0]} unless @_;

    my @old = @{$self->[0]};
    @{$self->[0]} = $self->_flatten(@_);
    return @old;
}

sub commands {
    my $self = shift;
    return @{$self->[1]} unless @_;

    my @old = @{$self->[1]};
    @{$self->[1]} = $self->_flatten(@_);
    return @old;
}

package PureMessage::Sieve::Command;

our @ISA = qw(PureMessage::Sieve::Node);

sub new {
    my($class, $id, @self) = @_;
    if (ref($id)) {
	unshift(@self, @$id);    # name/comment/line
    }
    else {
	unshift(@self, $id, "", undef);
    }
    bless \@self, $class;
}

sub name {
    my $self = shift;
    my $old = $self->[0];
    $self->[0] = shift if @_;
    return $old;
}

sub comment {
    my $self = shift;
    my $old = $self->[1];
    $self->[1] = shift if @_;
    return $old;
}

sub line {
    my $self = shift;
    my $old = $self->[2];
    $self->[2] = shift if @_;
    return $old;
}

sub args {
    my $self = shift;
    my @args = @$self;
    shift(@args);  # name
    shift(@args);  # comment
    shift(@args);  # line
    splice(@$self, 3, @$self, $self->_flatten(@_)) if @_;
    return @args;
}

package PureMessage::Sieve::Test;    our @ISA = qw(PureMessage::Sieve::Command);
package PureMessage::Sieve::Action;  our @ISA = qw(PureMessage::Sieve::Command);
package PureMessage::Sieve::If;      our @ISA = qw(PureMessage::Sieve::Command);

package PureMessage::Sieve::Cond;

our @ISA = qw(PureMessage::Sieve::Node);

sub new {
    my($class, @args) = @_;
    bless \@args, $class;
    return \@args;
}

sub test {
    my $self = shift;
    my $old = $self->[0];
    $self->[0] = shift if @_;
    $old;
}

sub comment {
    my $self = shift;
    my $old = $self->[1];
    $self->[1] = shift if @_;
    $old;
}

sub commands {
    my $self = shift;
    my @cmds = @$self;
    shift(@cmds);  # test
    shift(@cmds);  # comment
    splice (@$self, 2, @$self, $self->_flatten(@_)) if @_;
    return @cmds;
}

package PureMessage::Sieve::Testlist;

our @ISA = qw(PureMessage::Sieve::Node);

sub new {
    my $class = shift;
    my $self = bless [], $class;
    $self->tests(@_);
    $self;
}

sub tests {
    my $self = shift;
    my @old = @$self;
    if (@_) {
	@$self = $self->_flatten(@_);
    }
    return @old;
}

package PureMessage::Sieve::Stringlist;

our @ISA = qw(PureMessage::Sieve::Node);

sub new {
    my $class = shift;
    my $self = bless [], $class;
    $self->strings(@_);
    $self;
}

sub strings {
    my $self = shift;
    my @old = @$self;
    if (@_) {
	@$self = map { ref($_) ? $_ : PureMessage::Sieve::String->new($_) }
	             $self->_flatten(@_);
    }
    return @old;
}

package PureMessage::Sieve::Simple;

our @ISA = qw(PureMessage::Sieve::Node);

sub new {
    my($class, $self) = @_;
    bless \$self, $class
}

use overload '""' => sub { "${$_[0]}" }, fallback => 1;

package PureMessage::Sieve::String;  our @ISA = qw(PureMessage::Sieve::Simple);
package PureMessage::Sieve::Tag;     our @ISA = qw(PureMessage::Sieve::Simple);
package PureMessage::Sieve::Number;  our @ISA = qw(PureMessage::Sieve::Simple);

use overload '""' => sub {
    my $self = shift;
    my $num = $$self;
    if ($num == 0) {
	return $num;
    }
    elsif ($num % (1024 * 1024 * 1024) == 0) {
	return $num / (1024 * 1024 * 1024) . "G";
    }
    elsif ($num % (1024 * 1024) == 0) {
	return $num / (1024 * 1024) . "M";
    }
    elsif ($num % 1024 == 0) {
	return $num / 1024 . "K";
    }
    else {
	return $num;
    }
};

use overload '0+' => sub {
    my $self = shift;
    $$self;
};

use overload fallback => 1;

1;
