package PureMessage::Sieve::Parser;

use strict;
use Carp qw(croak);
use ActiveState::UTF8 qw();

require PureMessage::Sieve::Parser::Gen;

sub new {
    my $class = shift;
    bless {}, $class;
}

sub parse_file {
    my $self = shift;
    my $file = shift;
    my $parser = PureMessage::Sieve::Parser::Gen->new(make_file_lexer($file, @_), 0);
    return $parser->yyparse;
}

sub make_file_lexer {
    my $file = shift;
    
    open(my $fh, $file) || croak("Can't open file '$file': $!");
    
    my $line = "";
    my %keyword = map { $_ => 1 } qw(require if elsif else);
    my $com = "";
    my $trailing_com;

    return (

    # lexer
    sub {
	package PureMessage::Sieve::Parser::Gen;
	our($IDENT, $NUMBER, $STRING, $TAG);

      NEXT_TOKEN:
	while (1) {
	    unless (length($line)) {
		$line = <$fh>;		
	    }
	    return (0, undef) unless defined $line;
   	    die("Invalid UTF8 encoding in $file:\n\t$line\n") unless ActiveState::UTF8::is_valid_utf8($line);
	    $line =~ s/^\s+//;
	    if ($line =~ s/^\#\s?//) {
		$com .= $line unless $trailing_com;
		$line = "";
	    }
	    $trailing_com = 0;
	    last if length $line;
	};

	if ($line =~ s,^/\*,,) {
	    # /* ... */ comment
	    my $start_linenum = $.;
	    while ($line !~ s,.*?\*/,,) {
		$line = <$fh>;
		die "syntax error: Unterminated comment starting at $file line $start_linenum\n"
		    unless defined $line;
		die("Invalid UTF8 encoding in $file:\n\t$line\n") unless ActiveState::UTF8::is_valid_utf8($line);
	    }
	    goto NEXT_TOKEN;
	}

	my $prev_com = $com;
	$com = "";

	if ($line =~ s/^text:\s*(\#.*)?$//i) {
	    my @str;
	    local $_;
	    while (<$fh>) {
		die("Invalid UTF8 encoding in $file:\n\t$_\n") unless ActiveState::UTF8::is_valid_utf8($_);
		s/\r$//;
		last if /^\.$/;
		s/^\.\./\./;
		push(@str, $_);
	    }
	    return ($STRING, new_string(join("", @str)));
	}
	elsif ($line =~ s/^(:?[a-zA-Z]\w*)//) {
	    no strict 'refs';
	    my $token = $1;

	    # RFC 3028 says: "Tokens in the ASCII range are considered
            # case-insensitive", so we don't want to use lc() here.
	    $token =~ tr/A-Z/a-z/;

	    return (${uc($token)}, [$token, $prev_com, $.]) if $keyword{$token};
	    return ($TAG, new_tag($token)) if substr($token, 0, 1) eq ":";
	    return ($IDENT, [$token, $prev_com, $.]);
	}
	elsif ($line =~ s/^(\d+)([KMGkmg]?)//) {
	    my $num = $1;
	    if ($2) {
		$num *= 1024           if uc($2) eq "K";
		$num *= 1024*1024      if uc($2) eq "M";
		$num *= 1024*1024*1024 if uc($2) eq "G";
	    }
	    return ($NUMBER, new_number($num));
	}
	elsif ($line =~ s/^\"((?:[^\"\\]|\\[\"\\])*)\"//) {
	    my $str = $1;
	    $str =~ s/\\(.)/$1/g;
	    return ($STRING, new_string($str));
	}
	else {
	    my $c = substr($line, 0, 1, "");
	    $trailing_com++ if $c eq ";" || $c eq "}";
	    return (ord($c), $c);
	}
    },

    # error handler
    sub {
	croak("$_[0] at $file line $.\n");
    }

    );
}

1;
