package PureMessage::Sieve::Print;

use strict;
use ActiveState::Indenter;
use ActiveState::UTF8::File;

sub print {
    my $node = shift;
    my %opts;
    if (@_ == 1) {
	$opts{file} = shift;
    }
    else {
	%opts = @_;
    }

    my $fh = delete $opts{file} || *STDOUT;
    
    # removed the use of uUTF8::File as there is no longer a need to convert
    # latin characters. 
    $fh = ActiveState::Indenter->new($fh);
    for (qw(line_width indent_offset)) {
	my $v = delete $opts{$_};
	next unless defined $v;
	$fh->$_($v);
    }

    my $comment = delete $opts{comment};

    if (%opts && $^W) {
	require Carp;
	for (sort keys %opts) {
	    Carp::carp("Unrecognized option '$_'");
	}
    }

    if (defined $comment) {
	_print_comment($fh, $comment);
	$fh->print("\n");
    }

    my $type = $node->type;
    no strict 'refs';
    my $func = "_print_" . $type;
    &$func($fh, $node);
}

sub text {
    my $node = shift;
    my $accum = PureMessage::Sieve::Print::Accum->new;
    &print($node, file => $accum, @_);
    return $accum->value;
}

sub _print_script {
    my($fh, $script) = @_;
    if (my @requires = $script->requires) {
	$fh->print("require ");
	if (@requires > 1) {
	    _print_str_list($fh, @requires);
	}
	else {
	    _print_str($fh, $requires[0]);
	}
	$fh->print(";\n\n");
    }

    for my $cmd ($script->commands) {
	_print_cmd($fh, $cmd);
    }
}

sub _print_str {
    my($fh, $str) = @_;
    if ($str =~ /\n/) {
	$str =~ s/^\./../gm;
	$str .= "\n" unless $str =~ /\n\z/;
	$fh->over_abs(0);
	$fh->print(qq(text:\n$str.\n));
	$fh->back;
    }
    else {
	$str =~ s/\\/\\\\/g;
	$str =~ s/\"/\\\"/g;
	$fh->print(qq("$str"));
    }
}

sub _print_str_list {
    my $fh = shift;
    $fh->print("[");
    $fh->over_cur;
    while (@_) {
	_print_str($fh, $_[0]);
	shift;
	if (@_) {
	    $fh->print(",");
	    $fh->soft_space;
	}
    }
    $fh->back;
    $fh->print("]");
}

sub _print_comment {
    my($fh, $o) = @_;
    my $comment = ref($o) ? $o->comment : $o;
    if (length $comment) {
	$comment =~ s/^/\# /gm;
	$comment .= "\n" unless $comment =~ /\n\z/;
	$fh->print($comment);
    }
}

sub _print_cmd {
    my($fh, $cmd) = @_;
    
    my $name = $cmd->name;
    if ($name eq "if") {
	_print_if($fh, $cmd);
    }
    else {
	_print_comment($fh, $cmd);
	$fh->print("$name");
	_print_args($fh, $cmd->args);
	$fh->print(";\n");
    }
}

sub _print_if {
    my($fh, $if) = @_;
    my $kw = "if";
    for my $arg ($if->args) {
	if (my $test = $arg->test) {
	    $fh->print("}\n") if $kw ne "if";
	    _print_comment($fh, $arg);
	    $fh->print("$kw ");
	    $kw = "elsif";
	    my $start_line = $fh->line;
	    _print_test($fh, $test);
	    if ($fh->line == $start_line) {
		$fh->print(" {\n");
	    }
	    else {
		$fh->print("\n{\n");
	    }
	}
	else {
	    $fh->print("}\n");
	    _print_comment($fh, $arg);
	    $fh->print("else {\n");
	}
	$fh->over;
	for my $cmd ($arg->commands) {
	    _print_cmd($fh, $cmd);
	}
	$fh->back;
    }
    $fh->print("}\n");
}

sub _print_test {
    my($fh, $test) = @_;
    $fh->print($test->name);
    _print_args($fh, $test->args);
}

sub _print_args {
    my $fh = shift;
    return unless @_;

    my $type = $_[0]->type;
    if ($type eq "test") {
	$fh->print(" ");
	$fh->over_cur;
	_print_test($fh, $_[0]);
	$fh->back;
    }
    elsif ($type eq "testlist") {
	$fh->print("(") ;
	$fh->over_cur;
	my @tests = $_[0]->tests;
	my $i = 0;
	while (1) {
	    _print_test($fh, $tests[$i]);
	    last if ++$i == @tests;
	    $fh->print(",\n");
	}
	$fh->back;
	$fh->print(")");
    }
    else {
	$fh->print(" ");
	$fh->over_cur;
	my $i = 0;
	while (1) {
	    my $arg = $_[$i];
	    if (!ref($arg) || $arg->type eq "string") {
		_print_str($fh, $arg);
	    }
	    elsif ($arg->type eq "stringlist") {
		_print_str_list($fh, @$arg);
	    }
	    else {
		$fh->print("$arg");
	    }
	    last if ++$i == @_;
	    $fh->soft_space;
	}
	$fh->back;
    }
}

package PureMessage::Sieve::Print::Accum;

sub new {
    my $class = shift;
    bless [], $class;
}

sub print {
    my $self = shift;
    push(@$self, @_);
}

sub value {
    my $self = shift;
    return join("", @$self);
}

1;
